@@ -68,7 +68,7 @@ bcopy (from, to, len)
 #endif
 #endif
 
-char *version_string = "GNU sed version 1.16";
+char *version_string = "GNU sed version 1.17";
 
 /* Struct vector is used to describe a chunk of a compiled sed program.  
  * There is one vector for the main program, and one for each { } pair,
@@ -1258,7 +1258,6 @@ restart:
 
   for (cur_cmd = vec->v, n = vec->v_length; n; cur_cmd++, n--)
     {
-
     exe_loop:
       addr_matched = 0;
       if (cur_cmd->aflags & A1_MATCHED_BIT)
@@ -1389,9 +1388,11 @@ restart:
 
 	    n = line.length;
 	    tmp = line.text;
-	    /* Use --n so this'll skip the trailing newline */
-	    while (--n)
+	    while (n--)
 	      {
+		/* Skip the trailing newline, if there is one */
+		if (!n && (*tmp == '\n'))
+		  break;
 		if (width > 77)
 		  {
 		    width = 0;
@@ -1522,108 +1523,111 @@ restart:
 	  break;
 
 	case 's':
-	  if (!tmp.alloc)
-	    {
-	      tmp.alloc = 50;
-	      tmp.text = ck_malloc (50);
-	    }
-	  count = 0;
-	  start = 0;
-	  remain = line.length - 1;
-	  tmp.length = 0;
-	  rep = cur_cmd->x.cmd_regex.replacement;
-	  rep_end = rep + cur_cmd->x.cmd_regex.replace_length;
-
-	  while ((offset = re_search (cur_cmd->x.cmd_regex.regx,
-				      line.text,
-				      line.length - 1,
-				      start,
-				      remain,
-				      &regs)) >= 0)
-	    {
-	      count++;
-	      if (offset - start)
-		str_append (&tmp, line.text + start, offset - start);
-
-	      if (cur_cmd->x.cmd_regex.flags & S_NUM_BIT)
-		{
-		  if (count != cur_cmd->x.cmd_regex.numb)
-		    {
-		      int matched = regs.end[0] - regs.start[0];
-		      if (!matched) matched = 1;
-		      str_append (&tmp, line.text + regs.start[0], matched);
-		      start = (offset == regs.end[0]
-			       ? offset + 1 : regs.end[0]);
-		      remain = (line.length - 1) - start;
-		      continue;
-		    }
-		}
-
-	      for (rep_next = rep_cur = rep; rep_next < rep_end; rep_next++)
-		{
-		  if (*rep_next == '&')
-		    {
-		      if (rep_next - rep_cur)
-			str_append (&tmp, rep_cur, rep_next - rep_cur);
-		      str_append (&tmp, line.text + regs.start[0], regs.end[0] - regs.start[0]);
-		      rep_cur = rep_next + 1;
-		    }
-		  else if (*rep_next == '\\')
-		    {
-		      if (rep_next - rep_cur)
-			str_append (&tmp, rep_cur, rep_next - rep_cur);
-		      rep_next++;
-		      if (rep_next != rep_end)
-			{
-			  int n;
-
-			  if (*rep_next >= '0' && *rep_next <= '9')
-			    {
-			      n = *rep_next - '0';
-			      str_append (&tmp, line.text + regs.start[n], regs.end[n] - regs.start[n]);
-			    }
-			  else
-			    str_append (&tmp, rep_next, 1);
-			}
-		      rep_cur = rep_next + 1;
-		    }
-		}
-	      if (rep_next - rep_cur)
-		str_append (&tmp, rep_cur, rep_next - rep_cur);
-	      if (offset == regs.end[0])
-		{
-		  str_append (&tmp, line.text + offset, 1);
-		  ++regs.end[0];
-		}
-	      start = regs.end[0];
-
-	      remain = (line.length - 1) - start;
-	      if (remain < 0)
-		break;
-	      if (!(cur_cmd->x.cmd_regex.flags & S_GLOBAL_BIT))
-		break;
-	    }
-	  if (!count)
+	  {
+	    int trail_nl_p = line.text [line.length - 1] == '\n';
+	    if (!tmp.alloc)
+	      {
+		tmp.alloc = 50;
+		tmp.text = ck_malloc (50);
+	      }
+	    count = 0;
+	    start = 0;
+	    remain = line.length - trail_nl_p;
+	    tmp.length = 0;
+	    rep = cur_cmd->x.cmd_regex.replacement;
+	    rep_end = rep + cur_cmd->x.cmd_regex.replace_length;
+	    
+	    while ((offset = re_search (cur_cmd->x.cmd_regex.regx,
+					line.text,
+					line.length - trail_nl_p,
+					start,
+					remain,
+					&regs)) >= 0)
+	      {
+		count++;
+		if (offset - start)
+		  str_append (&tmp, line.text + start, offset - start);
+		
+		if (cur_cmd->x.cmd_regex.flags & S_NUM_BIT)
+		  {
+		    if (count != cur_cmd->x.cmd_regex.numb)
+		      {
+			int matched = regs.end[0] - regs.start[0];
+			if (!matched) matched = 1;
+			str_append (&tmp, line.text + regs.start[0], matched);
+			start = (offset == regs.end[0]
+				 ? offset + 1 : regs.end[0]);
+			remain = (line.length - trail_nl_p) - start;
+			continue;
+		      }
+		  }
+		
+		for (rep_next = rep_cur = rep; rep_next < rep_end; rep_next++)
+		  {
+		    if (*rep_next == '&')
+		      {
+			if (rep_next - rep_cur)
+			  str_append (&tmp, rep_cur, rep_next - rep_cur);
+			str_append (&tmp, line.text + regs.start[0], regs.end[0] - regs.start[0]);
+			rep_cur = rep_next + 1;
+		      }
+		    else if (*rep_next == '\\')
+		      {
+			if (rep_next - rep_cur)
+			  str_append (&tmp, rep_cur, rep_next - rep_cur);
+			rep_next++;
+			if (rep_next != rep_end)
+			  {
+			    int n;
+			    
+			    if (*rep_next >= '0' && *rep_next <= '9')
+			      {
+				n = *rep_next - '0';
+				str_append (&tmp, line.text + regs.start[n], regs.end[n] - regs.start[n]);
+			      }
+			    else
+			      str_append (&tmp, rep_next, 1);
+			  }
+			rep_cur = rep_next + 1;
+		      }
+		  }
+		if (rep_next - rep_cur)
+		  str_append (&tmp, rep_cur, rep_next - rep_cur);
+		if (offset == regs.end[0])
+		  {
+		    str_append (&tmp, line.text + offset, 1);
+		    ++regs.end[0];
+		  }
+		start = regs.end[0];
+		
+		remain = (line.length - trail_nl_p) - start;
+		if (remain < 0)
+		  break;
+		if (!(cur_cmd->x.cmd_regex.flags & S_GLOBAL_BIT))
+		  break;
+	      }
+	    if (!count)
+	      break;
+	    replaced = 1;
+	    str_append (&tmp, line.text + start, remain + trail_nl_p);
+	    t.text = line.text;
+	    t.length = line.length;
+	    t.alloc = line.alloc;
+	    line.text = tmp.text;
+	    line.length = tmp.length;
+	    line.alloc = tmp.alloc;
+	    tmp.text = t.text;
+	    tmp.length = t.length;
+	    tmp.alloc = t.alloc;
+	    if ((cur_cmd->x.cmd_regex.flags & S_WRITE_BIT)
+		&& cur_cmd->x.cmd_regex.wio_file)
+	      ck_fwrite (line.text, 1, line.length,
+			 cur_cmd->x.cmd_regex.wio_file);
+	    if (cur_cmd->x.cmd_regex.flags & S_PRINT_BIT)
+	      ck_fwrite (line.text, 1, line.length, stdout);
 	    break;
-	  replaced = 1;
-	  str_append (&tmp, line.text + start, remain + 1);
-	  t.text = line.text;
-	  t.length = line.length;
-	  t.alloc = line.alloc;
-	  line.text = tmp.text;
-	  line.length = tmp.length;
-	  line.alloc = tmp.alloc;
-	  tmp.text = t.text;
-	  tmp.length = t.length;
-	  tmp.alloc = t.alloc;
-	  if ((cur_cmd->x.cmd_regex.flags & S_WRITE_BIT)
-	      && cur_cmd->x.cmd_regex.wio_file)
-	    ck_fwrite (line.text, 1, line.length,
-		       cur_cmd->x.cmd_regex.wio_file);
-	  if (cur_cmd->x.cmd_regex.flags & S_PRINT_BIT)
-	    ck_fwrite (line.text, 1, line.length, stdout);
-	  break;
-
+	  }
+	    
 	case 't':
 	  if (replaced)
 	    {
@@ -1688,13 +1692,15 @@ match_address (addr)
       return (input_line_number == addr->addr_number);
 
     case addr_is_regex:
-      return (re_search (addr->addr_regex,
-			 line.text,
-			 line.length - 1,
-			 0,
-			 line.length - 1,
-			 (struct re_registers *) 0) >= 0) ? 1 : 0;
-
+      {
+	int trail_nl_p = line.text [line.length - 1] == '\n';
+	return (re_search (addr->addr_regex,
+			   line.text,
+			   line.length - trail_nl_p,
+			   0,
+			   line.length - trail_nl_p,
+			   (struct re_registers *) 0) >= 0) ? 1 : 0;
+      }
     case addr_is_last:
       return (input_EOF) ? 1 : 0;
 
